@@ -233,7 +233,6 @@ static xmlNodePtr
 _domExtractFrag(xmlNodePtr frag) {
     xmlNodePtr fraglist = frag->children;
     xmlNodePtr cur = fraglist;
-
     // detach fragment list
     frag->children = frag->last = NULL;
     while ( cur ){
@@ -256,13 +255,11 @@ domGetExternalSubset(xmlDocPtr self) {
 
 DLLEXPORT xmlDtdPtr
 domSetInternalSubset(xmlDocPtr self, xmlDtdPtr dtd) {
-    xmlDtdPtr ext_dtd = NULL;
     xmlDtdPtr int_dtd = NULL;
 
     assert(self != NULL);
 
     int_dtd = domGetInternalSubset(self);
-    ext_dtd = domGetExternalSubset(self);
 
     if (int_dtd == dtd) {
         return dtd;
@@ -272,29 +269,30 @@ domSetInternalSubset(xmlDocPtr self, xmlDtdPtr dtd) {
         domReleaseNode((xmlNodePtr) int_dtd);
     }
 
-    if (dtd->doc == NULL) {
-        xmlSetTreeDoc( (xmlNodePtr) dtd, self );
-    } else if ( dtd->doc != self ) {
-        XML6_FAIL(self, "moving DTDs between documents is not supported.");
-    }
+    if (dtd != NULL) {
+        xmlUnlinkNode((xmlNodePtr)dtd);
+        if (dtd->doc == NULL) {
+            xmlSetTreeDoc( (xmlNodePtr) dtd, self );
+        } else if ( dtd->doc != self ) {
+            XML6_FAIL(self, "moving DTDs between documents is not supported.");
+        }
 
-    if (dtd != NULL && ext_dtd == dtd) {
-        xmlUnlinkNode((xmlNodePtr) ext_dtd);
-        self->extSubset = NULL;
+        if (self->children == NULL)
+            xmlAddChild((xmlNodePtr) self, (xmlNodePtr) dtd);
+        else
+            xmlAddPrevSibling(self->children, (xmlNodePtr) dtd);
     }
     self->intSubset = dtd;
-    dtd->parent = self;
+
     return dtd;
 }
 
 DLLEXPORT xmlDtdPtr
 domSetExternalSubset(xmlDocPtr self, xmlDtdPtr dtd) {
     xmlDtdPtr ext_dtd = NULL;
-    xmlDtdPtr int_dtd = NULL;
 
     assert(self != NULL);
 
-    int_dtd = domGetInternalSubset(self);
     ext_dtd = domGetExternalSubset(self);
 
     if (ext_dtd == dtd) {
@@ -305,27 +303,52 @@ domSetExternalSubset(xmlDocPtr self, xmlDtdPtr dtd) {
         domReleaseNode((xmlNodePtr) ext_dtd);
     }
 
-    if ( dtd->doc == NULL ) {
-        xmlSetTreeDoc( (xmlNodePtr) dtd, self );
-    } else if (dtd->doc != self) {
-        XML6_FAIL(self, "moving DTDs between documents is not supported");
+    if (dtd != NULL) {
+        xmlUnlinkNode( (xmlNodePtr) dtd);
+        if ( dtd->doc == NULL ) {
+            xmlSetTreeDoc( (xmlNodePtr) dtd, self );
+        } else if (dtd->doc != self) {
+            XML6_FAIL(self, "moving DTDs between documents is not supported");
+        }
     }
 
-    if (dtd != NULL && int_dtd == dtd) {
-        xmlUnlinkNode( (xmlNodePtr) int_dtd);
-        self->intSubset = NULL;
-    }
     self->extSubset = dtd;
-    dtd->parent = NULL;
     return dtd;
 }
 
+DLLEXPORT xmlEntityPtr
+domGetEntityFromDtd(xmlDtdPtr dtd, xmlChar *name) {
+    xmlEntitiesTablePtr table;
+
+    if((dtd != NULL) && (dtd->entities != NULL)) {
+	table = (xmlEntitiesTablePtr) dtd->entities;
+	return((xmlEntityPtr) xmlHashLookup(table, name));
+    }
+    return(NULL);
+}
+
+DLLEXPORT xmlEntityPtr
+domGetParameterEntityFromDtd(xmlDtdPtr dtd, xmlChar *name) {
+    xmlEntitiesTablePtr table;
+
+    if ((dtd != NULL) && (dtd->pentities != NULL)) {
+	table = (xmlEntitiesTablePtr) dtd->pentities;
+	return((xmlEntityPtr) xmlHashLookup(table, name));
+    }
+    return(NULL);
+}
+
+
 static xmlNodePtr
 _domAssimulate(xmlNodePtr head, xmlNodePtr tail) {
     xmlNodePtr cur = head;
     while ( cur ) {
         /* we must reconcile all nodes in the fragment */
         if (cur->type == XML_DTD_NODE) {
+            if (_domIsDoc(cur->parent) == 0) {
+                xml6_warn("non-root DTD node found");
+                xmlUnlinkNode(cur);
+            }
             if (cur->doc && domGetExternalSubset(cur->doc) != (xmlDtdPtr)cur) {
                 domSetInternalSubset(cur->doc, (xmlDtdPtr) cur);
             }
@@ -356,63 +379,64 @@ _domAssimulate(xmlNodePtr head, xmlNodePtr tail) {
 static xmlNodePtr
 _domAddNodeToList(xmlNodePtr cur, xmlNodePtr leader, xmlNodePtr followup, xmlNodePtr *ptail) {
     xmlNodePtr head = NULL, tail = NULL, p = NULL, n = NULL;
-    if ( cur ) {
-        head = tail = cur;
-        if ( leader ) {
-            p = leader->parent;
-        }
-        else if ( followup ) {
-            p = followup->parent;
-        }
-        else {
-            return 0; /* can't insert */
-        }
+    assert(ptail != NULL);
 
-        if (leader && followup && p != followup->parent) {
-            warn("_domAddNodeToList(cur, prev, next, &frag) - 'prev' and 'next' have different parents");
-        }
-
-        if (cur->type == XML_DTD_NODE) {
-            xml6_warn("_domAddNodeToList(..) called on a DTD node");
-        }
-
-        if ( cur->type == XML_DOCUMENT_FRAG_NODE ) {
-            head = _domExtractFrag(cur);
-
-            n = head;
-            while ( n ){
-                n->parent = p;
-                n->doc = p->doc;
-                tail = n;
-                n = n->next;
-            }
-        }
-        else {
-            cur->parent = p;
-        }
-
-        if (head && tail && head != leader) {
-            if ( leader ) {
-                leader->next = head;
-                head->prev = leader;
-            }
-            else if ( p ) {
-                p->children = head;
-            }
-
-            if ( followup ) {
-                followup->prev = tail;
-                tail->next = followup;
-            }
-            else if ( p ) {
-                p->last = tail;
-            }
-        }
-        *ptail = tail;
-        return head;
+    if (cur == NULL) {
+        *ptail = NULL;
+        return NULL;
     }
-    *ptail = NULL;
-    return NULL;
+
+    head = tail = cur;
+    if ( leader ) {
+        p = leader->parent;
+    }
+    else if ( followup ) {
+        p = followup->parent;
+    }
+    else {
+        *ptail = NULL;
+        return NULL; /* can't insert */
+    }
+
+    if (leader && followup && p != followup->parent) {
+        warn("_domAddNodeToList(cur, prev, next, &frag) - 'prev' and 'next' have different parents");
+    }
+
+    if ( cur->type == XML_DOCUMENT_FRAG_NODE ) {
+        head = _domExtractFrag(cur);
+
+        n = head;
+        while ( n ){
+            n->parent = p;
+            n->doc = p->doc;
+            tail = n;
+            n = n->next;
+        }
+    }
+    else {
+        cur->parent = p;
+    }
+
+    if (head && tail && head != leader) {
+        if ( leader ) {
+            leader->next = head;
+            head->prev = leader;
+        }
+        else if ( p ) {
+            p->children = head;
+        }
+
+        if ( followup ) {
+            followup->prev = tail;
+            tail->next = followup;
+        }
+        else if ( p ) {
+            p->last = tail;
+        }
+    }
+
+    *ptail = tail;
+    return head;
 }
 
 static xmlNodePtr
@@ -421,17 +445,21 @@ _domSetDtd(xmlDocPtr doc, xmlDtdPtr dtd, xmlNodePtr old) {
     int replace_external = (old && old->type == XML_DTD_NODE ? (xmlDtdPtr)old : dtd) == ext_dtd;
 
     if (doc == NULL) XML6_FAIL(dtd, "DTD is not associated with a document");
-    if (_domIsDoc((xmlNodePtr)doc) == 0) {
+    if (_domIsDoc((xmlNodePtr)doc) == 0 || (dtd != NULL && dtd->type != XML_DTD_NODE)) {
         XML6_FAIL((xmlNodePtr)dtd, "SetDtd: HIERARCHY_REQUEST_ERR");
     }
-    if (old && old != (xmlNodePtr) dtd) xmlUnlinkNode(old);
 
-    if (replace_external) {
-        dtd = domSetExternalSubset(doc, dtd);
-    }
-    else {
-        dtd = domSetInternalSubset(doc, dtd);
+    if (old) xmlUnlinkNode(old);
+
+    if (dtd != (xmlDtdPtr)old) {
+        if (replace_external) {
+            dtd = domSetExternalSubset(doc, dtd);
+        }
+        else {
+            dtd = domSetInternalSubset(doc, dtd);
+        }
     }
+
     return (xmlNodePtr) dtd;
 }
 
@@ -522,6 +550,22 @@ domTestDocument(xmlNodePtr cur, xmlNodePtr refNode) {
     return 1;
 }
 
+static void _domScanEntry(void* value, int* refs, xmlChar* key) {
+    if (value != NULL && ((xmlNodePtr)value)->_private != NULL) {
+        (*refs)++;
+    }
+}
+
+static int
+_domScanHashForRefs(xmlHashTablePtr self) {
+    int refs = 0;
+    if (self != NULL) {
+        xmlHashScan(self, (xmlHashScanner)_domScanEntry, (void*) &refs);
+    }
+    return refs;
+}
+
+// Determine if there's any API references to the node or its descendants
 DLLEXPORT int
 domNodeIsReferenced(xmlNodePtr self) {
     xmlAttrPtr attr;
@@ -541,32 +585,47 @@ domNodeIsReferenced(xmlNodePtr self) {
     if (self->type == XML_ELEMENT_NODE) {
         // scan element attributes
         for (attr = self->properties; attr != NULL; attr = attr->next) {
-            if (attr->type == XML_ATTRIBUTE_NODE && attr->_private != NULL) {
+            void* ref = attr->type == XML_NAMESPACE_DECL
+                ? ((xmlNsPtr)attr)->_private
+                : attr->_private;
+            if (ref != NULL) {
                 return 1;
             }
         }
     }
-    else if (_domIsDoc(self)) {
-        // scan document dtds
-        xmlDocPtr doc = (xmlDocPtr) self;
-        if (doc->intSubset != NULL
-            && domNodeIsReferenced((xmlNodePtr)doc->intSubset)) {
+    else if (self->type == XML_DTD_NODE) {
+        xmlDtdPtr dtd = (xmlDtdPtr)self;
+        if (_domScanHashForRefs(dtd->elements)
+            || _domScanHashForRefs(dtd->attributes)
+            || _domScanHashForRefs(dtd->entities)) {
             return 1;
         }
-        if (doc->extSubset != NULL
-            && domNodeIsReferenced((xmlNodePtr)doc->extSubset)) {
+    }
+    else if (_domIsDoc(self)) {
+        // check external DtD
+        xmlDtdPtr dtd = ((xmlDocPtr) self)->extSubset;
+        if (dtd != NULL && domNodeIsReferenced((xmlNodePtr)dtd)) {
             return 1;
         }
     }
 
-    // scan siblings and children
-    for (cur = self; cur != NULL; cur = cur->next) {
-        xmlNodePtr kids = cur->children;
-        if (cur->_private != NULL
-            || (kids != NULL && domNodeIsReferenced(kids))) {
+    if (self->type == XML_ENTITY_DECL) {
+        xmlEntityPtr ent = (xmlEntityPtr) self;
+        if (ent->etype == XML_INTERNAL_PREDEFINED_ENTITY) {
+            // predefined entities are static. don't GC
             return 1;
         }
     }
+    else if (self->type != XML_ENTITY_REF_NODE) {
+        // scan siblings and children
+        for (cur = self; cur != NULL; cur = cur->next) {
+            xmlNodePtr kids = cur->children;
+            if (cur->_private != NULL
+                || (kids != NULL && domNodeIsReferenced(kids))) {
+                return 1;
+            }
+        }
+    }
 
     return 0;
 }
@@ -853,7 +912,7 @@ domAppendChild( xmlNodePtr self,
         xmlUnlinkNode( newChild );
     }
     else {
-        domImportNode( self->doc, newChild, 1, 0 );
+        newChild = domImportNode( self->doc, newChild, 1, 0 );
     }
 
     if ( self->children != NULL ) {
@@ -955,7 +1014,7 @@ domReplaceChild( xmlNodePtr self, xmlNodePtr new, xmlNodePtr old ) {
             xmlUnlinkNode( new );
         }
         else {
-            domImportNode( self->doc, new, 1, 1 );
+            new = domImportNode( self->doc, new, 1, 1 );
         }
 
         if( old == self->children && old == self->last ) {
@@ -1012,7 +1071,7 @@ domInsertBefore( xmlNodePtr self,
         xmlUnlinkNode( newChild );
     }
     else {
-        domImportNode( self->doc, newChild, 1, 0 );
+        newChild = domImportNode( self->doc, newChild, 1, 0 );
     }
 
     if ( refChild == NULL ) {
@@ -1080,7 +1139,7 @@ domReplaceNode( xmlNodePtr self, xmlNodePtr newNode ) {
                 xmlUnlinkNode( newNode );
             }
             else {
-                domImportNode( self->doc, newNode, 1, 0 );
+                newNode = domImportNode( self->doc, newNode, 1, 0 );
             }
 
             head = _domAddNodeToList( newNode, prev,  next, &tail );
