@@ -18,6 +18,13 @@ static xml6Ref ref_freed = {
     NULL, NULL, 0, 0, 0
 };
 
+static xmlMutexPtr _mutex = NULL;
+
+#ifdef DEBUG
+static int ref_current = 0;
+static int ref_total = 0;
+#endif
+
 DLLEXPORT void* xml6_ref_freed() {
     return (void *) &ref_freed;
 }
@@ -32,7 +39,6 @@ _ref_new(void) {
     return ref;
 }
 
-static xmlMutexPtr _mutex = NULL;
 DLLEXPORT void
 xml6_ref_init(void) {
     if (_mutex == NULL) {
@@ -56,6 +62,10 @@ xml6_ref_add(void** self_ptr) {
         if ( *self_ptr == NULL ) {
             self = _ref_new();
             *self_ptr = (void*) self;
+#ifdef DEBUG
+            ref_current++;
+            ref_total++;
+#endif
             init = 1;
         }
         xmlMutexUnlock(_mutex);
@@ -116,6 +126,11 @@ xml6_ref_remove(void** self_ptr, const char* name, void *obj) {
                     *self_ptr = NULL;
                     xmlFree((void*) self);
                     self = NULL;
+#ifdef DEBUG
+                    xmlMutexLock(_mutex);
+                    ref_current--;
+                    xmlMutexUnlock(_mutex);
+#endif
                     released = 1;
                 }
                 else {
@@ -212,3 +227,20 @@ xml6_ref_unlock(void* _self) {
     }
     return 0;
 }
+
+DLLEXPORT int
+xml6_ref_current(void) {
+#ifdef DEBUG
+    return ref_current;
+#else
+    return -1;
+#endif
+}
+DLLEXPORT int
+xml6_ref_total(void) {
+#ifdef DEBUG
+    return ref_total;
+#else
+    return -1;
+#endif
+}
